"""
Javascript generator
"""
import ast
from .. import (
    prepare_expr,
    remove_size_literal
)
from .base import JavaLikeGenerator

class JavascriptGenerator(JavaLikeGenerator):
    """ Javascript Generator """

    bd_attr_aliases = {
        'ZERO': 'ZERO()',
        'ONE': 'ONE()',
        'TEN': 'TEN()',
    }

    def generate(self):
        wr = self.writer

        wr.writeln("// the generated code requires a big decimal implementation, you can for example use `npm install bigdecimal`");
        wr.writeln("// const big = require('../node_modules/bigdecimal')");
        wr.writeln("// const BigDecimal = big.BigDecimal;");

        with self.writer.indent('function {}(params)'.format(self.class_name)):
            # create input and output vars
            for (comment, variables, is_input) in [
                    ('Input variables', self.parser.input_vars, True),
                    ('Output variables', self.parser.output_vars, False),
                    ('Internal variables', self.parser.internal_vars, False),
                ]:
                wr.nl()
                wr.writeln('/* {} */'.format(comment))
                for var in variables:
                    if var.comment is not None:
                        wr.nl()
                        self._write_comment(var.comment, False)
                    default = 'null'
                    if var.default:
                        val = remove_size_literal(var.default)
                        default = self.convert_to_js(val)
                    wr.writeln('this.{name} = {default};'.format(
                        name=var.name,
                        default=default
                    ))
                    if is_input:
                        with self.writer.indent('if (params["{}"] !== undefined)'.format(var.name)):
                            self.writer.writeln('this.set{cap}(params["{name}"]);'.format(
                                cap=var.name.capitalize(),
                                name=var.name
                            ))

        wr.nl()
        wr.writeln("/* Constants */")
        const_tpl = ("Object.defineProperty({self.class_name}, '{const.name}', "
                     "{{value: {converted}}});")
        for const in self.parser.constants:
            if const.comment is not None:
                wr.nl()
                self._write_comment(const.comment, False)
            value = const.value
            if const.type.endswith('[]'):
                value = '[{}]'.format(value[1:-1])
            converted = self.convert_to_js(value)
            wr.writeln(const_tpl.format(
                self=self,
                const=const,
                converted=converted
            ))
        # create setters for input vars
        var_tpl = '{self.class_name}.prototype.set{cap} = function(value)'
        for var in self.parser.input_vars:
            wr.nl()
            signature = var_tpl.format(
                self=self,
                cap=var.name.capitalize()
            )
            with wr.indent(signature):
                wr.writeln('this.{} = value;'.format(var.name))

        # create getters for output vars
        getter_tpl = '{self.class_name}.prototype.get{cap} = function()'
        for var in self.parser.output_vars:
            wr.nl()
            signature = getter_tpl.format(
                self=self,
                cap=var.name.capitalize(),
            )
            with wr.indent(signature):
                wr.writeln('return this.{};'.format(var.name))
        self._write_method(self.parser.main_method)
        for method in self.parser.methods:
            self._write_method(method)

        wr.writeln('module.exports = {};'.format(self.class_name));

    def _write_method(self, method):
        self.writer.nl()
        if method.comment:
            self._write_comment(method.comment, False)
        signature = '{self.class_name}.prototype.{name} = function()'.format(
            self=self,
            name=method.name
        )
        # actual method body
        with self.writer.indent(signature):
            self._write_stmt_body(method)

    def _conv_attribute(self, node):
        """ Override BaseGenerator in order to replace
            ZERO, ONE and TEN BigDecimal class constants
        """
        attr = node.attr
        val = self.to_code(node.value)
        accop = self.property_accessor_op
        if attr in self.bd_attr_aliases:
            attr = self.bd_attr_aliases[attr]
            accop = self.constant_accessor_op
        return val + [accop, attr]

    def convert_to_js(self, value):
        """ Converts java pseudo code into valid java code """
        tree = ast.parse(prepare_expr(value))
        node = tree.body[0].value
        return ''.join(self.to_code(node))



"""
LstGen generates php, python, go, or java code from PAP XML
"""
import re
import ast
from lxml import etree

# matches java-like long or double numbers, e.g.
# 123L or 3.12354D
NUMS_WITH_SIZE_RE = re.compile(r'([0-9]+)([LD]{1})')
J2PY_REPLACEMENTS = {
    'new BigDecimal': 'BigDecimalConstructor',
    '&&': 'and',
    '||': 'or'
}

def remove_size_literal(source):
    """ Remove LONG or DOUBLE literals from number code """
    if NUMS_WITH_SIZE_RE.findall(source):
        return NUMS_WITH_SIZE_RE.sub(r'\1', source)
    return source

def prepare_expr(source):
    """ Make java-like code python compatible """
    for (key, repl) in J2PY_REPLACEMENTS.items():
        source = source.replace(key, repl)
    return source.strip()

def parse_eval_stmt(source):
    """ parse java-like code in exec attribute
        of an EVAL tag
    """
    source = prepare_expr(source)
    tree = ast.parse(source)
    # an EVAL always contains a single assingment
    assign = tree.body[0]
    identifier = assign.targets[0].id
    return (identifier, assign.value)

def parse_condition_stmt(source):
    """ Parse java-like code in expr attribute
        of an IF tag.
    """
    source = prepare_expr(source)
    tree = ast.parse(source)
    return tree.body[0].value

def prev_comment(element):
    """ Return previous comment for an element """
    comment_elm = element.getprevious()
    comment = None
    if comment_elm is not None and comment_elm.tag is etree.Comment:
        comment = comment_elm.text
    return comment

class Comment(object):
    """ Represents an XML comment """

    def __init__(self, content):
        self.content = content

    @classmethod
    def from_element(cls, element):
        """ Create a Comment object from lxml Element """
        return cls(element.text)


class SimpleStmt(object):
    """ A simple statement without a body """

    def __init__(self, comment=None):
        self.comment = comment

    @classmethod
    def from_element(cls, element):
        """ Create a SimpleStmt object from lxml Element """
        return cls(prev_comment(element))

class StmtWithBody(SimpleStmt):
    """ Represents a statement with a body, e.g. <method>..body..</method> """

    def __init__(self, comment=None):
        super(StmtWithBody, self).__init__(comment)
        self.body = []

    def add(self, stmt):
        """ Adds a statement to body """
        self.body.append(stmt)

    @classmethod
    def from_element(cls, element):
        """ Creates a StmtWithBody object from lxml Element """
        stmt = super(StmtWithBody, cls).from_element(element)
        for child in element.getchildren():
            if child.tag in STMT_MAP:
                stmt.add(STMT_MAP[child.tag].from_element(child))
        return stmt

class EvalStmt(SimpleStmt):
    """ Represents an EVAL statement (<EVAL exec="FOO = BigDecimal.valueOf(10)"/>) """

    def __init__(self, comment=None, expr=None):
        super(EvalStmt, self).__init__(comment)
        self.expr = expr

    @classmethod
    def from_element(cls, element):
        """ Creates an EvalStmt object from lxml Element """
        stmt = super(EvalStmt, cls).from_element(element)
        stmt.expr = element.get('exec')
        return stmt


class ExecuteStmt(SimpleStmt):
    """ Represents an EXEC statement (<EXEC method="foo"/>) """

    def __init__(self, comment=None, method_name=None):
        super(ExecuteStmt, self).__init__(comment)
        self.method_name = method_name

    @classmethod
    def from_element(cls, element):
        """ Creates an ExecuteStmt object from lxml Element """
        stmt = super(ExecuteStmt, cls).from_element(element)
        stmt.method_name = element.get('method')
        return stmt


class ThenStmt(StmtWithBody):
    """ Represents a THEN statement (<THEN>...</THEN>)"""
    pass

class ElseStmt(StmtWithBody):
    """ Represents an ELSE statement (<ELSE>...</ELSE>)"""
    pass

class IfStmt(StmtWithBody):
    """ Represents an IF statement (<IF expr="FOO == 10">...</IF>)"""

    def __init__(self, comment=None, condition=None):
        super(IfStmt, self).__init__(comment)
        self.condition = condition

    @classmethod
    def from_element(cls, element):
        """ Creates an IfStmt object from lxml Element """
        stmt = super(IfStmt, cls).from_element(element)
        stmt.condition = element.get('expr')
        return stmt


class Method(StmtWithBody):
    """ Represents a METHOD (<METHOD name="MPARA">...</METHOD>)"""

    def __init__(self, comment=None, name=None):
        super(Method, self).__init__(comment)
        self.name = name

    @classmethod
    def from_element(cls, element):
        """ Creates a Method object from lxml Element """
        method = super(Method, cls).from_element(element)
        method.name = element.get('name')
        return method


class Prop(object):
    """ Base class for a variable or constant """

    def __init__(self, name, type):
        self.name = name
        self.type = type


class Var(Prop):
    """ Represents a variable definition:
        Input variable: <INPUT name"FOO" type="int" default="1"/>
        Output variable: <OUTPUT name"BAR" type="BigDecimal" default="new BigDecimal(0)"/>
        Internal variable: <INTERNAL name"BLAH" type="int"/>
    """

    def __init__(self, name, type, default=None, comment=None):
        super(Var, self).__init__(name, type)
        self.default = default
        self.comment = comment
        if not self.default:
            if self.type == 'int':
                self.default = '0'
            elif self.type == 'BigDecimal':
                self.default = 'new BigDecimal(0)'
            elif self.type == 'double':
                self.default = '0.0'
        else:
            if self.type == 'int' and '.' in self.default:
                self.default = str(int(float(self.default)))

    @classmethod
    def from_element(cls, element):
        """ Creates a Var object from lxml Element """
        comment = prev_comment(element)
        return cls(
            element.get('name'),
            element.get('type'),
            element.get('default'),
            comment
        )

class Const(Prop):
    """ Represents a constant definition:
    <CONSTANT name="TAB1" type="BigDecimal[]" value="{..}"/>
    """

    def __init__(self, name, type, value, comment=None):
        super(Const, self).__init__(name, type)
        self.value = value
        self.comment = comment

    @classmethod
    def from_element(cls, element):
        """ Creates a Const object from lxml Element """
        comment = prev_comment(element)
        return cls(
            element.get('name'),
            element.get('type'),
            element.get('value'),
            comment
        )


STMT_MAP = {
    'METHOD': Method,
    'IF': IfStmt,
    'THEN': ThenStmt,
    'ELSE': ElseStmt,
    'EVAL': EvalStmt,
    'COMMENT': Comment,
    'EXECUTE': ExecuteStmt,
    'MAIN': Method,
}

class PapParser(object):
    """ Parses a PAP XML file and prepares the contents
        for use by a language specific writer.
    """

    def __init__(self, tree_root):
        self.tree_root = tree_root
        self.internal_name = None
        self._input_vars = None
        self._output_vars = None
        self._internal_vars = None
        self._constants = None
        self._methods = None
        self.main_method = None
        self.parse()

    def parse(self):
        """ Parses the passed in XML doc root """
        self.repair_tree()
        self.internal_name = self.tree_root.get('name')
        main_element = self.tree_root.xpath('/PAP/METHODS/MAIN')[0]
        main_element.set('name', 'MAIN')
        self.main_method = Method.from_element(main_element)

    def repair_tree(self):
        """ Move <else></else> element outside of an <if></if> element
            back into <if> parent, see e.g. method MZTABFB from Lohnsteuer2010.xml
        """
        for elm in self.tree_root.xpath('//IF'):
            else_elm = elm.getnext()
            if else_elm is not None and else_elm.tag == 'ELSE':
                elm.getparent().remove(else_elm)
                elm.append(else_elm)

    @property
    def methods(self):
        """ Accessor for methods """
        if not self._methods:
            self._methods = []
            for method in self.tree_root.xpath('/PAP/METHODS/METHOD'):
                self._methods.append(Method.from_element(method))
        return self._methods

    def _get_vars(self, xpath):
        for var in self.tree_root.xpath(xpath):
            yield Var.from_element(var)

    @property
    def input_vars(self):
        """ Accessor for input variables """
        if not self._input_vars:
            self._input_vars = list(self._get_vars('/PAP/VARIABLES/INPUTS/INPUT'))
        return self._input_vars

    @property
    def output_vars(self):
        """ Accessor for output variables """
        if not self._output_vars:
            self._output_vars = list(self._get_vars('/PAP/VARIABLES/OUTPUTS/OUTPUT'))
        return self._output_vars

    @property
    def internal_vars(self):
        """ Accessor for internal variables """
        if not self._internal_vars:
            self._internal_vars = list(self._get_vars('/PAP/VARIABLES/INTERNALS/INTERNAL'))
        return self._internal_vars

    @property
    def constants(self):
        """ Accessor for constants variables """
        if not self._constants:
            self._constants = [
                Const.from_element(const)
                for const in self.tree_root.xpath('/PAP/CONSTANTS/CONSTANT')
            ]
        return self._constants

    @property
    def constant_names(self):
        """ Accessor for a set of constant names """
        return set(const.name for const in self.constants)



# coding: utf-8
"""
Base code generators module
"""
from .php import PhpGenerator
from .python import PythonGenerator
from .golang import GoLangGenerator
from .java import JavaGenerator
from .javascript import JavascriptGenerator

__all__ = [
    'PhpGenerator',
    'PythonGenerator',
    'GoLangGenerator',
    'JavaGenerator',
    'JavascriptGenerator'
]

GENERATORS = {
    'php': PhpGenerator,
    'python': PythonGenerator,
    'golang': GoLangGenerator,
    'java': JavaGenerator,
    'javascript': JavascriptGenerator
}



import decimal

class BigDecimal(decimal.Decimal):
    """ Compatibility class for decimal.Decimal """

    ROUND_DOWN = decimal.ROUND_DOWN
    ROUND_UP = decimal.ROUND_UP

    @classmethod
    def _mk_exp(cls, prec):
        return cls('0.' + '0' * prec)

    def divide(self, other, scale=None, rounding=None):
        if not scale and not rounding:
            return BigDecimal(self / other)
        if type(scale) is not int:
            raise ValueError("Expected integer value for scale")
        exp = BigDecimal._mk_exp(scale)
        return BigDecimal((self / other).quantize(exp, rounding=rounding))

    @classmethod
    def valueOf(cls, value):
        return cls(value)

    def multiply(self, other):
        return BigDecimal(self * other)

    def setScale(self, scale, rounding):
        exp = BigDecimal._mk_exp(scale)
        return BigDecimal(self.quantize(exp, rounding=rounding))

    def add(self, other):
        return BigDecimal(self + other)

    def subtract(self, other):
        return BigDecimal(self - other)

    def longValue(self):
        return int(self)

    def compareTo(self, other):
        return BigDecimal(self.compare(other))

BigDecimal.ZERO = BigDecimal(0)
BigDecimal.ONE = BigDecimal(1)
BigDecimal.TEN = BigDecimal(10)



# coding: utf-8
"""
Python writer
"""
import ast
import inspect

from ... import (
    prepare_expr,
    parse_eval_stmt,
    parse_condition_stmt,
    remove_size_literal
)
from ..base import BaseGenerator
from ... import (
    EvalStmt,
    IfStmt,
    ThenStmt,
    ElseStmt,
    ExecuteStmt,
)

from . import bd


class PythonGenerator(BaseGenerator):
    """ Python generator """


    bool_and = 'and'
    """ Boolean AND operator """

    bool_or = 'or'
    """ Boolean OR operator """

    bd_class_constructor = 'BigDecimal'
    """ Override BigDecimal class constructor """

    def __init__(self, parser, outfile, class_name=None, indent=None):
        super(PythonGenerator, self).__init__(
            parser,
            outfile,
            class_name,
            indent,
            block_chars=(':', None)
        )

    def _write_preamble(self):
        self.writer.writeln("# coding: utf-8")
        self.writer.nl()
        self.writer.writeln(inspect.getsource(bd))
        self.writer.nl()

    def generate(self):
        self._write_preamble()
        with self.writer.indent('class {}'.format(self.class_name)):
            for const in self.parser.constants:
                value = const.value
                if const.type.endswith('[]'):
                    value = '[{}]'.format(value[1:-1])
                converted = self.convert_to_python(value)
                self.writer.writeln('{const.name} = {converted}'.format(
                    const=const, converted=converted
                ))
                if const.comment is not None:
                    self._write_comment(const.comment, False)
                    self.writer.nl()
            self._write_constructor()
            # create setters for input vars
            for var in self.parser.input_vars:
                self.writer.nl()
                with self.writer.indent('def set{}(self, value)'.format(var.name.capitalize())):
                    if var.type == 'BigDecimal':
                        self.writer.writeln('self.{} = BigDecimal(value)'.format(var.name))
                    else:
                        self.writer.writeln('self.{} = value'.format(var.name))

            # create getters for output vars
            for var in self.parser.output_vars:
                self.writer.nl()
                with self.writer.indent('def get{}(self)'.format(var.name.capitalize())):
                    self.writer.writeln('return self.{}'.format(var.name))
            self._write_method(self.parser.main_method)
            for method in self.parser.methods:
                self._write_method(method)

    def _write_constructor(self):
        self.writer.nl()
        with self.writer.indent('def __init__(self, **kwargs)'):
            # initialize variables with default values
            for (variables, comment, is_input) in [
                    (self.parser.input_vars, 'input variables', True),
                    (self.parser.output_vars, 'output variables', False),
                    (self.parser.internal_vars, 'internal variables', False),
                ]:
                self.writer.writeln('# ' + comment)
                for var in variables:
                    if var.comment is not None:
                        self.writer.nl()
                        self._write_comment(var.comment, True)
                    self.writer.writeln('self.{name} = {value}'.format(
                        name=var.name,
                        value=self.convert_to_python(var.default)
                    ))
                    if is_input:
                        with self.writer.indent('if "{}" in kwargs'.format(var.name)):
                            self.writer.writeln('self.set{cap}(kwargs["{name}"])'.format(
                                cap=var.name.capitalize(),
                                name=var.name
                            ))
                self.writer.nl()

    def _write_method(self, method):
        self.writer.nl()
        with self.writer.indent('def {}(self)'.format(method.name)):
            if method.comment:
                self._write_comment(method.comment, False)
            self._write_stmt_body(method)

    def _write_comment(self, comment, simple=True):
        lines = comment.split("\n")
        if not simple:
            self.writer.writeln('"""')
        prefix = '# ' if simple else ''
        for line in lines:
            self.writer.writeln(u'{}{}'.format(prefix, line.strip()))
        if not simple:
            self.writer.writeln('"""')

    def _write_stmt_body(self, stmt):
        for part in stmt.body:
            if isinstance(part, EvalStmt):
                self.writer.writeln(self._convert_exec(part.expr))
            elif isinstance(part, ExecuteStmt):
                self.writer.writeln('self.{}()'.format(part.method_name))
            elif isinstance(part, IfStmt):
                self._write_if(part)
            elif isinstance(part, ElseStmt):
                self._write_else(part)
            elif isinstance(part, ThenStmt):
                # avoid empty if-statement body
                if part and part.body:
                    self._write_stmt_body(part)
                else:
                    self.writer.writeln('pass')

    def _write_if(self, stmt):
        converted = self._convert_if(stmt.condition)
        with self.writer.indent('if {}'.format(converted)):
            self._write_stmt_body(stmt)

    def _write_else(self, stmt):
        if not stmt.body:
            # avoid empty else stmts
            return
        self.writer.dec_indent()
        self.writer.writeln('else:')
        self.writer.inc_indent()
        self._write_stmt_body(stmt)

    def _convert_exec(self, expr):
        expr = remove_size_literal(expr)
        (var, parsed_stmt) = parse_eval_stmt(expr)
        ret = ['self.', var, ' = ']
        ret += self.to_code(parsed_stmt)
        return ''.join(ret)

    def _convert_if(self, expr):
        expr = remove_size_literal(expr)
        compare_stmt = parse_condition_stmt(expr)
        return ''.join(self.to_code(compare_stmt))

    def convert_to_python(self, value):
        """ Convert a java-like expression to valid python code """
        value = remove_size_literal(value)
        tree = ast.parse(prepare_expr(value))
        node = tree.body[0].value
        return ''.join(self.to_code(node))



# coding: utf-8
"""
AST to code converter base class
"""
import ast


class AstToCode(object):
    """ converts an AST to code """

    property_accessor_op = '.'
    """ Default property accessor operator for OOP """

    constant_accessor_op = '.'
    """ Default constant accessor operator for OOP """

    instance_var = 'self'
    """ Name of the implicit instance variable (if any) """

    add_op = '+'
    """ Addition operator (usually a "+") """

    unary_add_op = '+'
    """ Unary addition operator as in "+x" (usually a "+") """

    sub_op = '-'
    """ Subtraction operator (usually a "-") """

    unary_sub_op = '-'
    """ Unary subtraction operator as in "-x" (usually a "-") """

    mult_op = '*'
    """ Multiplication operator (usually a "*") """

    div_op = '/'
    """ Division operator (usually a "/") """

    list_subscription_parens = ('[', ']')
    """ Parenthesis to access a list/array member (subscription) """

    callable_exec_parens = ('(', ')')
    """ Parenthesis that denote the execution of a callable, as in "foo()" """

    call_args_delim = ', '
    """ Delimiter for function/method/callable arguments, usually a ", " """

    list_members_delim = ', '
    """ Delimiter for list/array members, usually a ", " """

    list_const_parens = ('[', ']')
    """ List/array constructor parenthesis, usually [], note that in case of PHP it's array() """

    bool_and = '&&'
    """ Boolean AND operator """

    bool_or = '||'
    """ Boolean OR operator """

    cmp_lt_op = '<'
    cmp_lte_op = '<='
    cmp_gt_op = '>'
    cmp_gte_op = '>='
    cmp_eq_op = '=='
    cmp_neq_op = '!='
    """ Comparison operators """

    bd_class = 'BigDecimal'
    """ BigDecimal class or its alias """

    bd_class_constructor = 'new BigDecimal'
    """ BigDecimal constructor statement, as in "new BigDecimal()" """

    allow_constants = True
    """ Allow usage of class constants (orf final static fields) """

    def __init__(self, parser, class_name=None):
        self.parser = parser
        self.class_name = class_name if class_name else self.parser.internal_name

    @property
    def inst_prefix(self):
        return '{}{}'.format(
            self.instance_var,
            self.property_accessor_op
        ) if self.instance_var else ''


    def _conv_attribute(self, node):
        return (
            self.to_code(node.value) +
            [self.property_accessor_op, node.attr]
        )

    def _conv_binop(self, node):
        return (
            self.to_code(node.left) +
            self.to_code(node.op) +
            self.to_code(node.right)
        )

    def _conv_name(self, node):
        if node.id == 'BigDecimal':
            return [self.bd_class]
        elif node.id == 'BigDecimalConstructor':
            return [self.bd_class_constructor]
        if self.allow_constants and node.id in self.parser.constant_names:
            return ['{}{}{}'.format(
                self.class_name,
                self.constant_accessor_op,
                node.id
            )]
        return ['{}{}'.format(self.inst_prefix, node.id)]

    def _conv_number(self, node):
        return [str(node.n)]

    def _conv_bool_op(self, node):
        vals = []
        op = self.bool_and
        if isinstance(node.op, ast.Or):
            op = self.bool_or
        op = ' {} '.format(op)
        for (idx, val) in enumerate(node.values):
            vals += self.to_code(val)
            if idx != len(node.values) - 1:
                vals.append(op)
        return vals

    def _conv_unary_op(self, node):
        op = self.unary_sub_op if isinstance(node.op, ast.USub) else self.unary_add_op
        return [op] + self.to_code(node.operand)

    def _conv_list_subscript(self, node):
        try:
            idx = node.slice.value
        except AttributeError:
            # python3.9 deprecated the ast.Index class for slices
            idx = node.slice
        return (
            self.to_code(node.value) +
            [self.list_subscription_parens[0]] +
            self.to_code(idx) +
            [self.list_subscription_parens[1]]
        )

    def _conv_call(self, node):
        args = []
        caller = self.to_code(node.func)
        for (idx, arg) in enumerate(node.args):
            args += self.to_code(arg)
            if idx != len(node.args) - 1:
                args.append(self.call_args_delim)
        return (
            caller +
            [self.callable_exec_parens[0]] +
            args +
            [self.callable_exec_parens[1]]
        )

    def _conv_comp(self, node):
        return (
            self.to_code(node.left) +
            self.to_code(node.ops[0]) +
            self.to_code(node.comparators[0])
        )

    def _conv_list(self, node):
        ret = [self.list_const_parens[0]]
        for (idx, elt) in enumerate(node.elts):
            ret += self.to_code(elt)
            if idx != len(node.elts) - 1:
                ret.append(self.list_members_delim)
        ret.append(self.list_const_parens[1])
        return ret

    def _conv_cmp_lt(self, node):
        return [' {} '.format(self.cmp_lt_op)]

    def _conv_cmp_lte(self, node):
        return [' {} '.format(self.cmp_lte_op)]

    def _conv_cmp_eq(self, node):
        return [' {} '.format(self.cmp_eq_op)]

    def _conv_cmp_gt(self, node):
        return [' {} '.format(self.cmp_gt_op)]

    def _conv_cmp_gte(self, node):
        return [' {} '.format(self.cmp_gte_op)]

    def _conv_cmp_neq(self, node):
        return [' {} '.format(self.cmp_neq_op)]

    def _conv_add_op(self, node):
        return [' {} '.format(self.add_op)]

    def _conv_sub_op(self, node):
        return [' {} '.format(self.sub_op)]

    def _conv_mult_op(self, node):
        return [' {} '.format(self.mult_op)]

    def _conv_div_op(self, node):
        return [' {} '.format(self.div_op)]

    def to_code(self, node):
        """ convert an AST node to its specific language representation """
        if isinstance(node, ast.Attribute):
            return self._conv_attribute(node)

        if isinstance(node, ast.BinOp):
            return self._conv_binop(node)

        if isinstance(node, ast.Name):
            return self._conv_name(node)

        if isinstance(node, ast.Num):
            return self._conv_number(node)

        if isinstance(node, ast.BoolOp):
            return self._conv_bool_op(node)

        if isinstance(node, ast.UnaryOp):
            return self._conv_unary_op(node)

        if isinstance(node, ast.Subscript):
            return self._conv_list_subscript(node)

        if isinstance(node, ast.Call):
            return self._conv_call(node)

        if isinstance(node, ast.List):
            return self._conv_list(node)

        if isinstance(node, ast.Compare):
            return self._conv_comp(node)

        if isinstance(node, ast.Lt):
            return self._conv_cmp_lt(node)

        if isinstance(node, ast.LtE):
            return self._conv_cmp_lte(node)

        if isinstance(node, ast.Eq):
            return self._conv_cmp_eq(node)

        if isinstance(node, ast.Gt):
            return self._conv_cmp_gt(node)

        if isinstance(node, ast.GtE):
            return self._conv_cmp_gte(node)

        if isinstance(node, ast.NotEq):
            return self._conv_cmp_neq(node)

        if isinstance(node, ast.Add):
            return self._conv_add_op(node)

        if isinstance(node, ast.Sub):
            return self._conv_sub_op(node)

        if isinstance(node, ast.Mult):
            return self._conv_mult_op(node)

        if isinstance(node, ast.Div):
            return self._conv_div_op(node)

        raise ValueError(u'Unknown AST element: {}'.format(node))



"""
PHP writer
BigDecimal functionality is provided by Brick/Math PHP module,
see https://github.com/brick/math
"""
import ast
import os

from ... import (
    prepare_expr,
    parse_eval_stmt,
    parse_condition_stmt,
    remove_size_literal
)
from ... import (
    EvalStmt,
    IfStmt,
    ThenStmt,
    ElseStmt,
    ExecuteStmt,
)

from .. base import BaseGenerator

class PhpGenerator(BaseGenerator):
    """ PHP Writer """

    property_accessor_op = '->'
    """ PHP property accessor operator """

    constant_accessor_op = '::'
    """ Default constant accessor operator for OOP """

    instance_var = '$this'
    """ Name of the implicit instance variable """

    list_const_parens = ('array(', ')')
    """ List/array constructor parenthesis, usually [], note that
        in case of PHP it's array()
    """

    allow_constants = False
    """ Disallow class constants, since we cannot have instances
        of BigDecimal as constant values
    """

    bd_attr_aliases = {
        'ZERO': 'zero()',
        'ONE': 'one()',
        'TEN': 'ten()',
    }

    bd_statics = (
        'valueOf',
        'ROUND_UP',
        'ROUND_DOWN',
    )

    def __init__(self, parser, outfile, class_name=None, indent=None, ns_name=''):
        super(PhpGenerator, self).__init__(parser, outfile, class_name, indent)
        self.ns_name = ns_name

    def _write_preamble(self):
        self.writer.writeln('<?php')
        self.writer.nl()
        if self.ns_name:
            self.writer.writeln('namespace {};'.format(self.ns_name))
            self.writer.nl()
        # add BigDecimal proxy class
        bd_php_path = os.path.join(os.path.dirname(__file__), 'BigDecimal.php')
        self.writer.nl()
        with open(bd_php_path) as fp:
            self.writer.writeln(fp.read())
        self.writer.nl()

    def generate(self):
        """ Generates PHP code """
        self._write_preamble()
        with self.writer.indent('class {}'.format(self.class_name)):
            self._write_comment('Internal variables', False)
            for var in self.parser.internal_vars:
                self._write_var(var)
            self.writer.nl()
            self._write_comment('Input variables', False)
            for var in self.parser.input_vars:
                self._write_var(var)
            self.writer.nl()
            self._write_comment('Output variables', False)
            for var in self.parser.output_vars:
                self._write_var(var)
            self._write_constructor()
            self._write_initializer()
            # create getters for output vars
            for var in self.parser.output_vars:
                self.writer.nl()
                signature = 'public function get{}()'.format(var.name.capitalize())
                with self.writer.indent(signature):
                    self.writer.writeln('return $this->{};'.format(var.name))

            # create setters for input vars
            for var in self.parser.input_vars:
                self.writer.nl()
                signature = 'public function set{}($value)'.format(var.name.capitalize())
                with self.writer.indent(signature):
                    if var.type == 'BigDecimal':
                        self.writer.writeln('$this->{} = BigDecimal::valueOf($value);'.format(var.name))
                    else:
                        self.writer.writeln('$this->{} = $value;'.format(var.name))

            self._write_method(self.parser.main_method, 'public')
            for method in self.parser.methods:
                self._write_method(method)

    def _write_constructor(self):
        self.writer.nl()
        with self.writer.indent('public function __construct()'):
            self.writer.writeln('$this->initialize();')

    def _write_initializer(self):
        self.writer.nl()
        with self.writer.indent('protected function initialize()'):
            # initialize variables with default values and set "constants"
            for (variables, comment) in [
                    (self.parser.internal_vars, 'Initialize internal variables'),
                    (self.parser.input_vars, 'Initialize input variables'),
                    (self.parser.output_vars, 'Initialize output variables')
                ]:
                self.writer.writeln('// ' + comment)
                for var in variables:
                    default = var.default
                    if not default:
                        if var.type == 'BigDecimal':
                            default = 'BigDecimal.valueOf(0)'
                        else:
                            default = '0'
                    self.writer.writeln('$this->{name} = {value};'.format(
                        name=var.name,
                        value=self.convert_to_php(default)
                    ))
                self.writer.nl()
            self.writer.writeln('// Initialize "constants"')
            for const in self.parser.constants:
                value = const.value
                if const.type.endswith('[]'):
                    value = '[{}]'.format(value[1:-1])
                converted = self.convert_to_php(value)
                self.writer.writeln('$this->{const.name} = {converted};'.format(
                    const=const, converted=converted
                ))

    def _write_var(self, var):
        if var.comment is not None:
            self.writer.nl()
            self._write_comment(var.comment, True)
        self.writer.writeln('private ${};'.format(var.name))

    def _write_method(self, method, visibility='protected'):
        self.writer.nl()
        if method.comment:
            self._write_comment(method.comment, False)
        signature = '{visibility} function {method.name}()'.format(
            visibility=visibility,
            method=method
        )
        with self.writer.indent(signature):
            self._write_stmt_body(method)

    def _write_comment(self, comment, simple=True):
        lines = comment.split("\n")
        if not simple:
            self.writer.writeln('/**')
        prefix = '//' if simple else ' *'
        for line in lines:
            self.writer.writeln(u'{} {}'.format(prefix, line.strip()))
        if not simple:
            self.writer.writeln(' */')

    def _write_stmt_body(self, stmt):
        for part in stmt.body:
            if isinstance(part, EvalStmt):
                self.writer.writeln(self._convert_exec(part.expr))
            elif isinstance(part, ExecuteStmt):
                self.writer.writeln('$this->{}();'.format(part.method_name))
            elif isinstance(part, IfStmt):
                self._write_if(part)
            elif isinstance(part, ElseStmt):
                self._write_else(part)
            elif isinstance(part, ThenStmt):
                self._write_stmt_body(part)

    def _write_if(self, stmt):
        converted = self._convert_if(stmt.condition)
        with self.writer.indent('if ({})'.format(converted)):
            self._write_stmt_body(stmt)

    def _write_else(self, stmt):
        if not stmt.body:
            # avoid empty else stmts
            return
        self.writer.dec_indent()
        self.writer.writeln('} else {')
        self.writer.inc_indent()
        self._write_stmt_body(stmt)

    def _convert_exec(self, expr):
        expr = remove_size_literal(expr)
        (var, parsed_stmt) = parse_eval_stmt(expr)
        ret = ['$this->', var, ' = ']
        ret += self.to_code(parsed_stmt)
        ret.append(";")
        return ''.join(ret)

    def _convert_if(self, expr):
        expr = remove_size_literal(expr)
        compare_stmt = parse_condition_stmt(expr)
        return ''.join(self.to_code(compare_stmt))

    def convert_to_php(self, value):
        value = remove_size_literal(value)
        tree = ast.parse(prepare_expr(value))
        node = tree.body[0].value
        return ''.join(self.to_code(node))

    def _conv_attribute(self, node):
        """ Override BaseGenerator in order to replace
            ZERO, ONE and TEN BigDecimal class constants
        """
        attr = node.attr
        val = self.to_code(node.value)
        accop = self.property_accessor_op
        if attr in self.bd_attr_aliases:
            attr = self.bd_attr_aliases[attr]
            accop = self.constant_accessor_op
        elif attr in self.bd_statics and val[-1] == 'BigDecimal':
            accop = self.constant_accessor_op
        return val + [accop, attr]




"""
Java generator
"""
import ast
from .. import prepare_expr
from .base import JavaLikeGenerator

class JavaGenerator(JavaLikeGenerator):
    """ Java Generator """

    def __init__(self, parser, outfile, class_name=None, indent=None, package_name='default'):
        super(JavaGenerator, self).__init__(parser, outfile, class_name, indent)
        self.package_name = package_name

    def generate(self):
        wr = self.writer
        if self.package_name:
            wr.writeln('package {};'.format(self.package_name))
            wr.nl()
        wr.writeln('import java.math.BigDecimal;')
        wr.nl()
        with self.writer.indent('public class {}'.format(self.class_name)):
            wr.writeln("/* Constants */")
            for const in self.parser.constants:
                if const.comment is not None:
                    wr.nl()
                    self._write_comment(const.comment, False)
                value = const.value
                is_list_initializer = False
                if const.type.endswith('[]'):
                    value = '[{}]'.format(value[1:-1])
                    is_list_initializer = True
                converted = self.convert_to_java(value)
                if is_list_initializer:
                    # convert python-style list back to java
                    converted = '{{{}}}'.format(converted[1:-1])
                wr.writeln('protected final static {const.type} {const.name} = {converted};'.format(
                    const=const,
                    converted=converted
                ))
            # create input and output vars
            for (comment, variables) in [
                    ('Input variables', self.parser.input_vars),
                    ('Output variables', self.parser.output_vars),
                    ('Internal variables', self.parser.internal_vars),
                ]:
                wr.nl()
                wr.writeln('/* {} */'.format(comment))
                for var in variables:
                    if var.comment is not None:
                        wr.nl()
                        self._write_comment(var.comment, False)
                    wr.writeln('protected {var.type} {var.name} = {var.default};'.format(var=var))
            # create setters for input vars
            for var in self.parser.input_vars:
                wr.nl()
                signature = 'public void set{cap}({type} value)'.format(
                    cap=var.name.capitalize(),
                    type=var.type
                )
                with wr.indent(signature):
                    wr.writeln('this.{} = value;'.format(var.name))

            # create getters for output vars
            for var in self.parser.output_vars:
                wr.nl()
                signature = 'public {type} get{cap}()'.format(
                    cap=var.name.capitalize(),
                    type=var.type
                )
                with wr.indent(signature):
                    wr.writeln('return this.{};'.format(var.name))
            self._write_method(self.parser.main_method, 'public')
            for method in self.parser.methods:
                self._write_method(method)
        wr.nl()

    def _write_method(self, method, visibility='protected'):
        self.writer.nl()
        if method.comment:
            self._write_comment(method.comment, False)
        signature = '{visibility} void {name}()'.format(
            visibility=visibility,
            name=method.name
        )
        # actual method body
        with self.writer.indent(signature):
            self._write_stmt_body(method)

    def convert_to_java(self, value):
        """ Converts java pseudo code into valid java code """
        tree = ast.parse(prepare_expr(value))
        node = tree.body[0].value
        return ''.join(self.to_code(node))



# coding: utf-8
"""
LstGen command line utils
"""
import sys
import os
import argparse
import codecs
from importlib.metadata import version

from lxml import etree

from . import PapParser
from . import pap
from .generators import GENERATORS

LANGUAGES = sorted(GENERATORS.keys())

def error(msg):
    """ Show an error message and exit """
    sys.stderr.write(msg)
    sys.stderr.write("\n")
    sys.stderr.flush()
    sys.exit(1)

def get_version():
    return version('lstgen')

def main():
    """ main lstgen function """
    parser = argparse.ArgumentParser(
        description='Erzeugt validen Code für die Lohnsteuerberechung aus PAP XML'
    )
    parser.add_argument(
        '-V', '--version',
        action='version',
        version='LstGen Version {}'.format(get_version()))
    parser.add_argument(
        '-l', '--lang',
        dest='lang',
        choices=LANGUAGES,
        metavar='LANG',
        help='Programmiersprache ({})'.format(', '.join(LANGUAGES)))
    parser.add_argument(
        '-p', '--pap-version',
        dest='pap_version',
        choices=pap.PAP_RESOURCES.keys(),
        metavar='PAP',
        help=('PAP Version (z.B. "2016"), falls kein PAP XML Pfad '
              'angegeben wurde, siehe auch Option --pap-versions')
    )
    parser.add_argument(
        '-x', '--pap-xml',
        dest='pap_xml_path',
        metavar='XML_PATH',
        help='Pfad zur PAP XML, falls keine PAP Version ausgewählt wurde.'
    )
    parser.add_argument(
        '--pap-versions',
        dest='show_pap_versions',
        action='store_true',
        default=False,
        help='Gibt die Liste der verfügbaren PAP Version aus.'
    )
    parser.add_argument(
        '--outfile',
        dest='outfile',
        metavar='OUTFILE',
        help='Ausgabedatei, default: STDOUT'
    )
    parser.add_argument(
        '--class-name',
        dest='class_name',
        metavar='CLASS_NAME',
        help="Name der generierten Klasse, standardmässig wird der Name der PAP Elements verwendet",
    )
    parser.add_argument(
        '--indent',
        dest='indent',
        default=u'    ',
        help='Zeichenkette mit der eingerückt wird, default: 4 Leerzeichen'
    )
    parser.add_argument(
        '--java-package-name',
        dest='java_package',
        metavar='JAVA_PACKAGE',
        help="Package-Name (falls LANG=java), standardmässig wird 'default' verwendet",
    )
    parser.add_argument(
        '--go-package-name',
        dest='go_package',
        metavar='GO_PACKAGE',
        help="Package-Name (falls LANG=golang), standardmässig wird 'tax' verwendet",
    )
    parser.add_argument(
        '--php-ns',
        dest='php_ns',
        metavar='PHP_NAMESPACE',
        help="Namespace-Name (falls LANG=php)",
    )

    args = parser.parse_args()

    if args.show_pap_versions:
        print("Verfügbare PAP Versionen:")
        for version in pap.PAP_RESOURCES.keys():
            print(version)
        sys.exit(0)

    if (not args.pap_version and not args.pap_xml_path) or (args.pap_version and args.pap_xml_path):
        error("Bitte entweder eine PAP Version (-p) oder Pfad zu einer PAP XML Datei angeben (-x).")

    xml_content = None
    if args.pap_xml_path:
        if not os.path.isfile(args.pap_xml_path):
            error("Kann Datei nicht öffnen: '{}'".format(args.pap_xml_path))
        else:
            xml_content = open(args.pap_xml_path, 'r').read()

    if args.pap_version:
        try:
            xml_content = pap.get_pap_xml(args.pap_version)
        except ValueError as err:
            error(err)

    lang = args.lang.lower()
    pap_parser = PapParser(etree.fromstring(xml_content))
    if not args.outfile:
        outfp = sys.stdout
    else:
        outfp = codecs.open(args.outfile, 'w+', encoding='utf-8')
    with outfp:
        gen_class = GENERATORS.get(lang)
        if not gen_class:
            error('No matching writer for langugage {}'.format(lang))

        if lang == 'php':
            generator = gen_class(
                pap_parser,
                outfp,
                class_name=args.class_name,
                ns_name=args.php_ns,
                indent=args.indent,
            )
        elif lang == 'python':
            generator = gen_class(
                pap_parser,
                outfp,
                class_name=args.class_name,
                indent=args.indent
            )
        elif lang == 'golang':
            generator = gen_class(
                pap_parser,
                outfp,
                class_name=args.class_name,
                package_name=args.go_package,
                indent=args.indent
            )    
        elif lang == 'java':
            generator = gen_class(
                pap_parser,
                outfp,
                class_name=args.class_name,
                package_name=args.java_package,
                indent=args.indent
            )
        elif lang == 'javascript':
            generator = gen_class(
                pap_parser,
                outfp,
                class_name=args.class_name,
                indent=args.indent
            )
        generator.generate()



# coding: utf-8
"""
PAP XML and remote API URLs
"""
from collections import (
    namedtuple,
    OrderedDict
)
import requests
from lxml import etree

PAP_BASE_URL = 'https://www.bmf-steuerrechner.de'

PapResource = namedtuple('PapResource', ('remote_service_path', 'pap_xml_path'))

PAP_RESOURCES = OrderedDict((
    ('2006_1', PapResource(
        '/interface/2006Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2006Big.xml.xhtml'
    )),
    ('2007_1', PapResource(
        '/interface/2007Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2007Big.xml.xhtml'
    )),
    ('2008_1', PapResource(
        '/interface/2008Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2008Big.xml.xhtml'
    )),
    ('2009_1', PapResource(
        '/interface/2009Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2009Big.xml.xhtml'
    )),
    ('2010_1', PapResource(
        '/interface/2010Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2010Big.xml.xhtml'
    )),
    ('2011_1', PapResource(
        '/interface/2011bisNovVersion1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2011BisNovember.xml.xhtml'
    )),
    ('2011_12', PapResource(
        '/interface/2011DezVersion1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2011Dezember.xml.xhtml'
    )),
    ('2012_1', PapResource(
        '/interface/2012Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2012.xml.xhtml'
    )),
    ('2013_1', PapResource(
        '/interface/2013Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2013.xml.xhtml'
    )),
    ('2014_1', PapResource(
        '/interface/2014Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2014.xml.xhtml'
    )),
    ('2015_1', PapResource(
        '/interface/2015bisNovVersion1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2015BisNovember.xml.xhtml'
    )),
    ('2015_12', PapResource(
        '/interface/2015DezVersion1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2015Dezember.xml.xhtml'
    )),
    ('2016_1', PapResource(
        '/interface/2016Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2016.xml.xhtml'
    )),
    ('2017_1', PapResource(
        '/interface/2017Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2017.xml.xhtml'
    )),
    ('2018_1', PapResource(
        '/interface/2018Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2018.xml.xhtml'
    )),
    ('2019_1', PapResource(
        '/interface/2019Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2019.xml.xhtml'
    )),
    ('2020_1', PapResource(
        '/interface/2020Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2020.xml.xhtml'
    )),
    ('2021_1', PapResource(
        '/interface/2021Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2021.xml.xhtml'
    )),
    ('2022_1', PapResource(
        '/interface/2022Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2022.xml.xhtml'
    )),
    ('2023_2', PapResource(
        '/interface/2023Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2023.xml.xhtml'
    )),
    ('2023_3', PapResource(
        '/interface/2023AbJuliVersion1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2023AbJuli.xml.xhtml'
    )),
    ('2024_1', PapResource(
        '/interface/2024Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2024.xml.xhtml'
    )),
    ('2024_2', PapResource(
        '/interface/2024DezemberVersion1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2024Dezember.xml.xhtml'
    )),
    ('2025_1', PapResource(
        '/interface/2025Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2025.xml.xhtml'
    )),
    ('2026_1', PapResource(
        '/interface/2026Version1.xhtml',
        '/javax.faces.resource/daten/xmls/Lohnsteuer2026.xml.xhtml'
    )),
))


def get_pap_xml(pap_resource_name):
    """ Fetch PAP XML from bmf-steuerrechner.de and return
        it as a string.
    """
    resource = PAP_RESOURCES.get(pap_resource_name)
    if not resource:
        raise ValueError("Invalid PAP resource name: {}".format(pap_resource_name))
    url = "{}{}".format(PAP_BASE_URL, resource.pap_xml_path)
    response = requests.get(url, verify=False) # FIXME bmf-steuerrechner.de has a weak certificate
    return response.content

def call_pap_service(pap_resource_name, input_vars):
    """ Call PAP remote service (REST XML API) and
        return a dict with results
    """
    resource = PAP_RESOURCES.get(pap_resource_name)
    if not resource:
        raise ValueError("Invalid PAP resource name: {}".format(pap_resource_name))
    url = "{}{}".format(PAP_BASE_URL, resource.remote_service_path)
    response = requests.get(url, input_vars, verify=False)
    tree = etree.fromstring(response.content)
    ret = dict()
    for elm in tree.xpath('/lohnsteuer/ausgaben/ausgabe'):
        ret[elm.get('name')] = elm.get('value')
    return ret



"""
GoLang generator
"""
import ast
import os
from ... import (
    parse_eval_stmt,
    prepare_expr
)
from ... import (
    EvalStmt,
    IfStmt,
    ThenStmt,
    ElseStmt,
    ExecuteStmt,
)
from .. base import JavaLikeGenerator

class GoLangGenerator(JavaLikeGenerator):
    """ GoLang Generator """
    bd_class = 'BigDecimal'    
    bd_class_constructor = 'NewFromInt'
    list_const_parens = ('{', '}')
    stmt_separator = ''
    allow_constants = False
    instance_var = 't'

    def __init__(self, parser, outfile, class_name=None, indent=None, package_name='default'):
        super(GoLangGenerator, self).__init__(parser, outfile, class_name, indent)
        self.package_name = package_name
        self.type_map = {}   # Keep a list of known variable types

    def _write_preamble(self):        
        # add BigDecimal proxy class
        bd_golang_path = os.path.join(os.path.dirname(__file__), 'BigDecimal.go')        
        with open(bd_golang_path) as fp:
            self.writer.writeln(fp.read())
        

    def generate(self):
        wr = self.writer
        self._write_comment("This file is automatically generated by LstGen2, do not edit!", False)
        wr.writeln('package {}'.format(self.package_name or 'tax'))        
        wr.nl()
        self._write_preamble()
        wr.nl()

        # Define the model as a struct.
        with self.writer.indent('type {} struct'.format(self.class_name)):
            wr.writeln("// ------------------------ Constants -------------------------")
            for const in self.parser.constants:
                if const.comment is not None:
                    wr.nl()
                    self._write_comment(const.comment, False)

                if const.type.endswith('[]'):
                    const.value = '[{}]'.format(const.value[1:-1])

                thetype = self._convert_vartype(const.type)
                wr.writeln('{const.name} {thetype}'.format(
                    const=const,
                    thetype=thetype
                ))
                self.type_map[const.name] = thetype

            for (comment, variables) in [
                    ('Input variables', self.parser.input_vars),
                    ('Output variables', self.parser.output_vars),
                    ('Internal variables', self.parser.internal_vars),
                ]:
                wr.nl()
                wr.writeln('// {}'.format(comment))

                wr.writeln("// ------------------------ Variables -------------------------")
                for var in variables:
                    if var.comment is not None:
                        wr.nl()
                        self._write_comment(var.comment, False)
                    vartype = self._convert_vartype(var.type)
                    wr.writeln('{} {}'.format(var.name, vartype))
                    self.type_map[var.name] = vartype

        # Define the New() method that also sets default values.
        with self.writer.indent('func New{cls}() *{cls}'.format(cls=self.class_name)):
            with self.writer.indent('return &{}'.format(self.class_name)):
                wr.writeln("// ------------------------ Constants -------------------------")
                for const in self.parser.constants:
                    if const.comment is not None:
                        wr.nl()
                        self._write_comment(const.comment, False)

                    if const.type.endswith('[]'):
                        const.value = '[{}]'.format(const.value[1:-1])

                    value = self.convert_to_go(const.value)
                    wr.writeln('{const.name}: {value},'.format(
                        const=const,
                        value=value
                    ))

                wr.nl()
                wr.writeln("// ------------------------ Variables -------------------------")
                for (comment, variables) in [
                        ('Input variables', self.parser.input_vars),
                        ('Output variables', self.parser.output_vars),
                        ('Internal variables', self.parser.internal_vars),
                    ]:

                    for var in variables:
                        if var.default is None:
                            continue

                        if var.comment is not None:
                            wr.nl()
                            self._write_comment(var.comment, False)

                        value = self.convert_to_go(var.default)
                        wr.writeln('{}: {},'.format(var.name, value))


        # create setters for input vars
        for var in self.parser.input_vars:
            wr.nl()
            signature = 'func ({instance} *{cls}) Set{cap}(value {type})'.format(
                instance=self.instance_var,
                cls=self.class_name,
                cap=var.name.capitalize(),
                type=self._convert_vartype(var.type)
            )
            with wr.indent(signature):
                wr.writeln('{instance}.{name} = value'.format(
                    instance=self.instance_var,
                    name=var.name
                ))

        # create getters for output vars
        for var in self.parser.output_vars:
            wr.nl()
            signature = 'func ({instance} *{cls}) Get{cap}() {type}'.format(
                instance=self.instance_var,
                cls=self.class_name,
                cap=var.name.capitalize(),
                type=self._convert_vartype(var.type)
            )
            with wr.indent(signature):
                wr.writeln('return {instance}.{name}'.format(
                    instance=self.instance_var,
                    name=var.name
                ))

        self._write_method(self.parser.main_method)
        for method in self.parser.methods:
            self._write_method(method)
        wr.nl()

    def _convert_vartype(self, vartype):
        return {
            'BigDecimal.ZERO': 'NewFromInt(0)',
            'BigDecimal[]': '[]'+self.bd_class,
            'BigDecimal': ''+self.bd_class,
            'int': 'int64',
            'double': 'float64',
        }[vartype]

    def _conv_list(self, node):
        res = super(GoLangGenerator, self)._conv_list(node)
        return ['[]' + self.bd_class] + res

    def _conv_attribute(self, node):
        clsmethod = False
        if node.attr == 'valueOf':
            node.attr = 'NewFromInt'
            clsmethod = True
        elif node.attr == 'ZERO':
            node.attr = 'NewFromInt(0)'
            clsmethod = True
        elif node.attr == 'ONE':
            node.attr = 'NewFromInt(1)'
            clsmethod = True
        elif node.attr == 'TEN':
            node.attr = 'NewFromInt(10)'
            clsmethod = True
        elif node.attr == 'longValue':
            node.attr = 'IntPart'
        elif node.attr == 'add':
            node.attr = 'Add'
        elif node.attr == 'subtract':
            node.attr = 'Subtract'
        elif node.attr == 'multiply':
            node.attr = 'Multiply'
        elif node.attr == 'divide':
            node.attr = 'Divide'
        elif node.attr == 'compareTo':
            node.attr = 'CompareTo'
        elif node.attr == 'setScale':
            node.attr = 'SetScale'
        elif node.attr in ('ROUND_UP', 'ROUND_DOWN'):
            return [node.attr]
        else:
            raise NotImplementedError("Unmapped attribute {}".format(node.attr))

        if clsmethod:
            return [node.attr]
        return (
            self.to_code(node.value) +
            [self.property_accessor_op, node.attr]
        )

    def _get_decimal_constructor_from_node(self, node):
        try:
            int(str(node.n))
            return 'NewFromInt'
        except ValueError:
            return 'NewFromFloat'
        except AttributeError: # If node.nodes[0] is not a value, but e.g. a name or a call
            # Unfortunately it's not possible to support all cases due to the strict
            # typing of Go. However, we are "lucky":
            if node.__class__.__name__ == 'Name':
                # For now, all variable assignments that create new values happen to be integers.
                rtype = self.type_map.get(node.id)
                if rtype == 'int64':
                    return 'NewFromInt'
                elif rtype == 'float64':
                    return 'NewFromFloat'
                raise NotImplementedError(node.id+' has unsupported type '+repr(rtype))
            elif node.__class__.__name__ == 'Call':
                # The node is a function call. Luckily so far, all calls return integers...
                # luckily, the only function calls in assignments are calls to Decimal methods,
                # which we can ensure to return int64
                return 'NewFromInt'
            elif node.__class__.__name__ == 'Constant':
                try:
                    int(str(node.value))
                    return 'NewFromInt'
                except ValueError:
                    return 'NewFromFloat'
            elif node.__class__.__name__ == 'BinOp':
                # For math operations, the type depends on the operators.
                return self._get_decimal_constructor_from_node(node.left)
        raise NotImplementedError("unsupported node type {}".format(
            node.__class__.__name__
        ))

    def _conv_call(self, node):
        caller = self.to_code(node.func)

        # Fix calls to the Round function (Go's Decimal doesn't accept a rounding parameter,
        # so we need to map it to another function here.)
        if caller[-1] == 'Round' and len(node.args) > 1:
            if node.args[-1] == 'ROUND_DOWN':
                caller[-1] = 'RoundCash'
            node.args = node.args[:1]

        # Similarly for the Div function.
        elif caller[-1] == 'Div' and len(node.args) > 1:
            # FIXME: Div does not support ROUND_DOWN :-(
            caller[-1] = 'Divide'
            node.args = node.args[:2]

        # Fix integer initialization-
        elif caller[-1] == 'NewFromInt':
            caller[-1] = self._get_decimal_constructor_from_node(node.args[0])

        args = []
        for (idx, arg) in enumerate(node.args):
            args += self.to_code(arg)
            if idx != len(node.args) - 1:
                args.append(self.call_args_delim)

        return (
            caller +
            [self.callable_exec_parens[0]] +
            args +
            [self.callable_exec_parens[1]]
        )

    def _write_method(self, method):
        self.writer.nl()
        if method.comment:
            self._write_comment(method.comment, False)
        signature = 'func ({instance} *{cls}) {name}()'.format(
            instance=self.instance_var,
            cls=self.class_name,
            name=method.name
        )
        # actual method body
        with self.writer.indent(signature):
            self._write_stmt_body(method)

    def convert_to_go(self, value):
        """ Converts java pseudo code into valid go code """
        tree = ast.parse(prepare_expr(value))
        node = tree.body[0].value
        return ''.join(self.to_code(node))



# coding: utf-8
"""
Base writers module
"""
from contextlib import contextmanager

from .ast2code import AstToCode

from .. import (
    parse_eval_stmt,
    parse_condition_stmt
)
from .. import (
    EvalStmt,
    IfStmt,
    ThenStmt,
    ElseStmt,
    ExecuteStmt,
)


class Writer(object):
    """ Basic writer class used to write structured
        code with indentation to a file-like object.
    """

    default_indent_str = ' ' * 4
    """ Default indentation string (4 spaces)"""

    def __init__(self, outfile, indent_str=None, block_chars=(' {', '}')):
        self.outfile = outfile
        self.block_chars = block_chars
        self.indent_str = indent_str if indent_str is not None else self.default_indent_str
        self.indent_lvl = 0

    def generate(self):
        raise Exception("Implement me!")

    def write(self, content, do_indent=True):
        """ Write content to outfile with optional indentation """
        if not content:
            return
        if do_indent:
            self.outfile.write(self.indent_str * self.indent_lvl)
        self.outfile.write(content)

    def writeln(self, content, do_indent=True):
        """ Write content to outfile finishing with a new-line """
        self.write(content, do_indent)
        self.nl()

    def nl(self, num=1, do_indent=False):
        """ Write a new-line to outfile """
        self.write("\n" * num, do_indent)

    def inc_indent(self, lvl=1):
        """ Increase indentation level """
        self.indent_lvl += lvl

    def dec_indent(self, lvl=1):
        """ Decrease indentation level """
        if self.indent_lvl > 0:
            self.indent_lvl -= lvl

    @contextmanager
    def indent(self, preamble, lvl=1):
        """ Allows creating visual indented blocks
            that produce actual code blocks:

            with x.indent('def foo()'):
                x.writeln('print("Hello World!"))
                x.writeln('return 1')
        """
        self.writeln('{}{}'.format(preamble, self.block_chars[0]))
        self.inc_indent(lvl)
        yield
        self.dec_indent(lvl)
        if self.block_chars[1]:
            self.writeln(self.block_chars[1])


class BaseGenerator(AstToCode):
    """ Base code generator class """

    def __init__(self, parser, outfile, class_name=None, indent=None, block_chars=(' {', '}')):
        super(BaseGenerator, self).__init__(parser, class_name)
        self.writer = Writer(outfile, indent, block_chars)

    def generate(self):
        raise NotImplementedError("Implement me!")


class JavaLikeGenerator(BaseGenerator):
    """ Base generator for java-like code syntax """

    stmt_separator = ';'
    """ Line endings (usually a ";" or empty) """

    instance_var = 'this'
    """ Name of the implicit instance variable (if any) """

    def __init__(self, parser, outfile, class_name=None, indent=None, package_name='default'):
        super(JavaLikeGenerator, self).__init__(parser, outfile, class_name, indent)
        self.package_name = package_name

    def _write_comment(self, comment, simple=True):
        lines = comment.split("\n")
        if not simple:
            self.writer.writeln('/**')
        prefix = '// ' if simple else ' * '
        for line in lines:
            self.writer.writeln(u'{}{}'.format(prefix, line.strip()))
        if not simple:
            self.writer.writeln(' */')

    def _write_stmt_body(self, stmt):
        for part in stmt.body:
            if isinstance(part, EvalStmt):
                self.writer.writeln(self._convert_exec(part.expr))
            elif isinstance(part, ExecuteStmt):
                self.writer.writeln('{}.{}(){}'.format(
                    self.instance_var,
                    part.method_name,
                    self.stmt_separator
                ))
            elif isinstance(part, IfStmt):
                self._write_if(part)
            elif isinstance(part, ElseStmt):
                self._write_else(part)
            elif isinstance(part, ThenStmt):
                self._write_stmt_body(part)

    def _write_if(self, stmt):
        converted = self._convert_if(stmt.condition)
        with self.writer.indent('if ({})'.format(converted)):
            self._write_stmt_body(stmt)

    def _write_else(self, stmt):
        if not stmt.body:
            # avoid empty else stmts
            return
        self.writer.dec_indent()
        self.writer.writeln('} else {')
        self.writer.inc_indent()
        self._write_stmt_body(stmt)

    def _convert_exec(self, expr):
        (var, parsed_stmt) = parse_eval_stmt(expr)
        ret = [self.instance_var, '.', var, ' = ']
        ret += self.to_code(parsed_stmt)
        ret.append(self.stmt_separator)
        return ''.join(ret)

    def _convert_if(self, expr):
        compare_stmt = parse_condition_stmt(expr)
        return ''.join(self.to_code(compare_stmt))
